import sys
import pytest

import numpy as np
import pandas as pd

from typing import Any
from typing import Dict
from pytest import approx
from pathlib import Path
from pandas.core.frame import DataFrame

sys.path.append(Path(__file__).parent.resolve()) # type: ignore

from daspi.statistics.confidence import *


source = Path(__file__).parent/'data'
KW_READ: Dict[str, Any] = dict(sep=';', index_col=0)

df_dist10: DataFrame = pd.read_csv(
    source/f'dists_10-samples.csv', skiprows=1, nrows=10, **KW_READ)
df_valid10: DataFrame = pd.read_csv(
    source/f'dists_10-samples.csv', skiprows=14, **KW_READ)
df_dist25: DataFrame = pd.read_csv(
    source/f'dists_25-samples.csv', skiprows=1, nrows=25, **KW_READ)
df_valid25: DataFrame = pd.read_csv(
    source/f'dists_25-samples.csv', skiprows=29, **KW_READ)

# source data contains 8 decimal places
rel_central: float = 1e-7
# The calculated confidence intervals in Minitab have only 4 decimal places
rel_interval: float = 1e-3


def test_confidence_to_alpha() -> None:
    level = 0.95
    alpha = confidence_to_alpha(level, two_sided=False, n_groups=1)
    assert alpha == approx(0.05)

    alpha = confidence_to_alpha(level, two_sided=True, n_groups=1)
    assert alpha == approx(0.025)

    alpha = confidence_to_alpha(level, two_sided=False, n_groups=5)
    assert alpha == approx(0.01)

    level = 0.9
    alpha = confidence_to_alpha(level, two_sided=True, n_groups=2)
    assert alpha == approx(0.025)
    
    with pytest.raises(AssertionError, match=r'\d+ not in \(0, 1\)'):
        confidence_to_alpha(5)

def test_mean_ci() -> None:
    rows = ['mean', 'mean_95ci_low', 'mean_95ci_upp']
    
    size = 10
    for dist in df_dist10.columns:
        x_bar, ci_low, ci_upp = mean_ci(df_dist10[dist], level=.95)
        _x_bar, _ci_low, _ci_upp = df_valid10[dist][rows]
        assert x_bar == approx(_x_bar, rel=rel_central)
        assert ci_low == approx(_ci_low, rel=rel_interval)
        assert ci_upp == approx(_ci_upp, rel=rel_interval)
    
    size = 25
    for dist in df_dist25.columns:
        x_bar, ci_low, ci_upp = mean_ci(df_dist25[dist], level=.95)
        _x_bar, _ci_low, _ci_upp = df_valid25[dist][rows]
        assert x_bar == approx(_x_bar, rel=rel_central)
        assert ci_low == approx(_ci_low, rel=rel_interval)
        assert ci_upp == approx(_ci_upp, rel=rel_interval)

def test_stdev_ci() -> None:
    rows = ['std', 'std_95ci_low', 'std_95ci_upp']
    
    size = 10
    for dist in df_dist10.columns:
        std, ci_low, ci_upp = stdev_ci(df_dist10[dist], level=.95)
        _std, _ci_low, _ci_upp = df_valid10[dist][rows]
        assert std == approx(_std, rel=rel_central)
        assert ci_low == approx(_ci_low, rel=rel_interval)
        assert ci_upp == approx(_ci_upp, rel=rel_interval)
    
    size = 25
    for dist in df_dist25.columns:
        std, ci_low, ci_upp = stdev_ci(df_dist25[dist], level=.95)
        _std, _ci_low, _ci_upp = df_valid25[dist][rows]
        assert std == approx(_std, rel=rel_central)
        assert ci_low == approx(_ci_low, rel=rel_interval)
        assert ci_upp == approx(_ci_upp, rel=rel_interval)

def test_bonferroni_ci() -> None:
    data = pd.DataFrame({
        'target': np.random.normal(0, 1, 100),
        'feature': np.random.choice(['A', 'B', 'C'], 100)
    })
    # Test case 1: Check if the output DataFrame has the expected columns
    result = bonferroni_ci(data, 'target', 'feature')
    assert all([c in result.columns for c in ['midpoint', 'lower', 'upper', 'feature']])

    # Test case 2: Check if the confidence intervals are within the correct range
    assert (result['lower'] <= result['midpoint']).all()
    assert (result['midpoint'] <= result['upper']).all()

    # Test case 3: Check if the number of groups matches the expected value
    assert len(result) == len(data['feature'].unique())

def test_bonferroni_ci_multiple_features() -> None:
    data = pd.DataFrame({
        'target': np.random.normal(0, 1, 100),
        'feature1': np.random.choice(['A', 'B', 'C'], 100),
        'feature2': np.random.choice(['X', 'Y', 'Z'], 100)
    })

    # Test case 1: Check if the output DataFrame has the expected columns
    result = bonferroni_ci(data, 'target', ['feature1', 'feature2'])
    assert all(col in result.columns for col in ['midpoint', 'lower', 'upper', 'feature1', 'feature2'])

    # Test case 2: Check if the confidence intervals are within the correct range
    assert (result['lower'] <= result['midpoint']).all()
    assert (result['midpoint'] <= result['upper']).all()

    # Test case 3: Check if the number of groups matches the expected value
    assert len(result) == len(data['feature1'].unique()) * len(data['feature2'].unique())

def test_confidence_to_alpha_edge_cases() -> None:
    # Test two-sided with n_groups > 1
    alpha = confidence_to_alpha(0.95, two_sided=True, n_groups=2)
    assert alpha == approx(0.0125)
    # Test one-sided with n_groups > 1
    alpha = confidence_to_alpha(0.95, two_sided=False, n_groups=2)
    assert alpha == approx(0.025)
    # Test invalid level
    with pytest.raises(AssertionError):
        confidence_to_alpha(-0.1)
    with pytest.raises(AssertionError):
        confidence_to_alpha(1.1)

def test_mean_ci_edge_cases() -> None:
    # Test with small sample
    arr = np.array([1.0, 2.0, 3.0])
    x_bar, ci_low, ci_upp = mean_ci(arr, level=0.99)
    assert x_bar == approx(np.mean(arr))
    assert ci_low < x_bar < ci_upp
    # Test with all identical values
    arr = np.ones(10)
    x_bar, ci_low, ci_upp = mean_ci(arr, level=0.95)
    assert x_bar == approx(1.0)
    assert pd.isna(ci_low)
    assert pd.isna(ci_upp)

def test_stdev_ci_edge_cases() -> None:
    # Test with small sample
    arr = np.array([1.0, 2.0, 3.0])
    std, ci_low, ci_upp = stdev_ci(arr, level=0.99)
    assert std == approx(np.std(arr, ddof=1))
    assert ci_low <= std <= ci_upp
    # Test with all identical values
    arr = np.ones(10)
    std, ci_low, ci_upp = stdev_ci(arr, level=0.95)
    assert std == approx(0.0)
    assert ci_low == approx(0.0)
    assert ci_upp == approx(0.0)

def test_bonferroni_ci_edge_cases() -> None:
    # Test with empty DataFrame
    df = pd.DataFrame({'target': [], 'feature': []})
    result = bonferroni_ci(df, 'target', 'feature')
    assert result.empty
    # Test with one group
    df = pd.DataFrame({'target': np.random.normal(0, 1, 10), 'feature': ['A']*10})
    result = bonferroni_ci(df, 'target', 'feature')
    assert len(result) == 1
    # Test with missing values
    df = pd.DataFrame({'target': [1, 2, np.nan, 4], 'feature': ['A', 'A', 'B', 'B']})
    result = bonferroni_ci(df, 'target', 'feature')
    assert all([c in result.columns for c in ['midpoint', 'lower', 'upper', 'feature']])

def test_proportion_ci_basic() -> None:
    # Typical case
    p, ci_low, ci_upp = proportion_ci(50, 100, level=0.95)
    assert 0 <= p <= 1
    assert ci_low <= p <= ci_upp
    # Edge case: zero events
    p, ci_low, ci_upp = proportion_ci(0, 100, level=0.95)
    assert p == 0
    assert ci_low <= p <= ci_upp
    # Edge case: all events
    p, ci_low, ci_upp = proportion_ci(100, 100, level=0.95)
    assert p == 1
    assert ci_low <= p <= ci_upp

def test_delta_mean_ci_basic() -> None:
    # Compare two samples
    a = np.random.normal(0, 1, 30)
    b = np.random.normal(1, 1, 30)
    delta, ci_low, ci_upp = delta_mean_ci(a, b, level=0.95)
    assert ci_low <= delta <= ci_upp

def test_delta_variance_ci_basic() -> None:
    a = np.random.normal(0, 1, 30)
    b = np.random.normal(0, 2, 30)
    delta, ci_low, ci_upp = delta_variance_ci(a, b, level=0.95)
    assert ci_low <= delta <= ci_upp

def test_delta_stdev_ci_basic() -> None:
    a = np.random.normal(0, 1, 30)
    b = np.random.normal(0, 2, 30)
    delta, ci_low, ci_upp = delta_stdev_ci(a, b, level=0.95)
    assert ci_low <= delta <= ci_upp

def test_delta_proportion_ci_basic() -> None:
    # Compare two proportions
    delta, ci_low, ci_upp = delta_proportions_ci(30, 100, 50, 100, level=0.95)
    assert ci_low <= delta <= ci_upp
    # Edge case: zero events in both
    delta, ci_low, ci_upp = delta_proportions_ci(0, 100, 0, 100, level=0.95)
    assert ci_low <= delta <= ci_upp

def test_delta_mean_ci_edge_cases() -> None:
    # Identical samples
    a = np.ones(10)
    b = np.ones(10)
    delta, ci_low, ci_upp = delta_mean_ci(a, b, level=0.95)
    assert delta == approx(0.0)
    assert ci_low == approx(0.0)
    assert ci_upp == approx(0.0)

def test_delta_stdev_ci_edge_cases() -> None:
    a = np.ones(10)
    b = np.ones(10)
    delta, ci_low, ci_upp = delta_stdev_ci(a, b, level=0.95)
    assert delta == approx(0.0)
    assert ci_low == approx(0.0)
    assert ci_upp == approx(0.0)

def test_delta_proportion_ci_edge_cases() -> None:
    # Identical proportions
    delta, ci_low, ci_upp = delta_proportions_ci(50, 100, 50, 100, level=0.95)
    assert delta == approx(0.0)
    assert ci_low <= delta <= ci_upp



import warnings

import numpy as np
import pandas as pd
import statsmodels.api as sm

from typing import Any
from typing import Literal
from typing import Iterable
from collections import defaultdict
from pandas.core.frame import DataFrame
from pandas.core.series import Series
from statsmodels.regression.linear_model import RegressionResultsWrapper

from .convert import get_term_name

from ..constants import ANOVA


__all__ = [
    'uniques',
    'terms_effect',
    'variance_inflation_factor',
    'anova_table',
    'terms_probability',]


def uniques(seq: Iterable) -> list[Any]:
    """Get a list of unique elements from a sequence while preserving 
    the original order.

    Parameters
    ----------
    seq : Iterable
        The input sequence.

    Returns
    -------
    List[Any]
        A list of unique elements from the input sequence, preserving 
        the original order.

    Notes
    -----
    This function is based on the 'uniqify' algorithm by Peter Bengtsson.
    Source: https://www.peterbe.com/plog/uniqifiers-benchmark

    Examples
    --------
    ```python
    sequence = [1, 2, 3, 2, 1, 4, 5, 4]
    unique_elements = uniques(sequence)
    print(unique_elements)
    ```
    
    ```console
    [1, 2, 3, 4, 5]
    ```
    """
    seen = set()
    return [x for x in seq if not (x in seen or seen.add(x))]

def terms_effect(model: RegressionResultsWrapper) -> Series:
    """Calculates the impact of each term on the target. The
    effects are described as absolute number of the parameter 
    coefficients devided by its standard error.

    Parameters
    ----------
    model : RegressionResultsWrapper
        Statsmodels regression results of fitted model.
    
    Returns
    -------
    Series
        A pandas Series containing the effects of each feature
        on the target variable.
    """
    params: Series = model.params
    se: Series = model.scale
    names_map = {n: get_term_name(n) for n in params.index}
    se = (model.bse
        .rename(index=names_map)
        .pow(2)
        .groupby(level=0)
        .sum()
        .pow(1/2))
    params = (params
        .rename(index=names_map)
        .groupby(level=0)
        .sum())
    effects = params.abs() / se
    effects = effects[uniques(names_map.values())]
    effects.name = ANOVA.EFFECTS
    effects.index.name = ANOVA.FEATURES
    return effects

def variance_inflation_factor(
        model: RegressionResultsWrapper,
        threshold: int = 5,
        generalized: bool = True) -> DataFrame:
    """Calculate the variance inflation factor (VIF) and the generalized
    variance inflation factor (GVIF) for each predictor variable in the
    fitted model.
        
    This function takes a regression model as input and returns a 
    DataFrame containing the VIF, GVIF (= VIF^(1/2*dof)), threshold for 
    GVIF, collinearity status and calculation kind for each predictor 
    variable in the model. The VIF and GVIF are measures of 
    multicollinearity, which can help identify variables that are highly
    correlated with each other.
        
    Parameters
    ----------
    model : RegressionResultsWrapper
        The regression model to analyze.
    threshold : int, optional
        The threshold for deciding whether a predictor is collinear.
        Common values are 5 and 10. By default 5.
    generalized : bool, optional
        Whether to calculate the generalized VIF or not, by default True.
    
    Returns
    -------
    DataFrame
        A DataFrame containing the VIF, GVIF, threshold, collinearity
        status and performed method for each predictor variable in the 
        model.
    
    Notes
    -----
    The VIF tells us: The degree to which the standard error of the 
    predictor is increased due to the predictor's correlation with the 
    other predictors in the model. VIF values greater than 10 
    (or, Tolerance values less than 0.10) corresponding to a multiple 
    correlation of 0.95 indicates a multicollinearity may be a problem 
    (Hair Jr, JF, Anderson, RE, Tatham, RL and Black, WC, 1998). Fox and 
    Weisberg also comment that the straightforward VIF can't be used if 
    there are variables with more than one degree of freedom (e.g. 
    polynomial and other contrasts relating to categorical variables 
    with more than two levels) and recommend using the gvif function 
    (generalized variance inflation factor) in the car package in R in 
    these cases. gvif is the square root of the VIF for individual 
    predictors and thus can be used equivalently. More generally 
    generalized variance-inflation factors consist of the VIF corrected 
    by the number of degrees of freedom (df) of the predictor variable: 
    GVIF = VIF[1/(2*df)] and may be compared to thresholds of 
    10[1/(2*df)] to assess collinearity using the stepVIF (source code: 
    https://github.com/cran/car/blob/master/R/vif.R) function in R.

    source: https://imaging.mrc-cbu.cam.ac.uk/statswiki/FAQ/Collinearity

    See Also:
    https://www.rdocumentation.org/packages/car/versions/3.1-2/topics/vif
    https://www.statsmodels.org/dev/generated/statsmodels.stats.outliers_influence.variance_inflation_factor.html
    """
    exog = (pd
        .DataFrame(
            model.model.data.exog.copy(),
            columns=list(map(get_term_name, model.model.data.param_names)))
        .drop_duplicates()
        .reset_index(drop=True))
    assert exog.shape[1] > 1, (
        'To calculate VIFs, at least two predictor variables must be present.')
    
    terms = uniques(exog.columns)
    vifs = pd.DataFrame(index=terms, columns=ANOVA.VIF_COLNAMES)
    det = np.linalg.det
    interaction_order_terms = defaultdict(list)
    for term in terms:
        interaction_order_terms[len(term.split(ANOVA.SEP))].append(term)
    for order, _terms in interaction_order_terms.items():
        for term in _terms:
            Y = exog[term]
            XY = exog[_terms]
            if ANOVA.INTERCEPT in XY and term != ANOVA.INTERCEPT:
                XY = XY.drop(ANOVA.INTERCEPT, axis=1)
            X = XY.drop(term, axis=1)
            vif = None
            method = f'single_order-{order}_term'
            if X.empty:
                vif = 1.0
            elif Y.ndim == 2 and X.ndim == 2 and generalized:
                vif = (det(Y.corr()) * det(X.corr())) / det(XY.corr())
                method = 'generalized'
            if pd.isna(vif):
                y = Y.iloc[:, 0] if Y.ndim == 2 else Y
                X = sm.add_constant(X) if term != ANOVA.INTERCEPT else X
                vif = 1 / (1 - sm.OLS(y, X).fit().rsquared)
                method = 'R_squared'
            vifs.loc[term, ANOVA.VIF] = vif
            vifs.loc[term, 'Method'] = method

    vifs['DF'] = [exog[t].shape[1] if exog[t].ndim == 2 else 1 for t in terms]
    vifs['GVIF'] = vifs['VIF']**(1/(2*vifs['DF']))
    vifs['Threshold'] = threshold**(1/(2*vifs['DF']))
    vifs['Collinear'] = vifs['GVIF'] >= vifs['Threshold']
    return vifs

def anova_table(
        model: RegressionResultsWrapper,
        typ: Literal['I', 'II', 'III']
        ) -> DataFrame:
        """Perform an analysis of variance (ANOVA) on the fitted model.

        Parameters
        ----------
        model : RegressionResultsWrapper
            A fitted regression model of the `statsmodels` package.
        typ : Literal['I', 'II', 'III'], optional
            The type of ANOVA to perform. Default is 'III', see notes
            for more informations about the types.
            - '' : If no or an invalid type is specified, Type-II is 
            used if the model has no significant interactions. 
            Otherwise, Type-III is used for hierarchical models and 
            Type-I is used for non-hierarchical models.
            - 'I' : Type I sum of squares ANOVA.
            - 'II' : Type II sum of squares ANOVA.
            - 'III' : Type III sum of squares ANOVA.

        Returns
        -------
        DataFrame
            The ANOVA table as DataFrame containing the following
            columns:
            - DF : Degrees of freedom for model terms.
            - SS : Sum of squares for model terms.
            - F : F statistic value for significance of adding model
            terms.
            - p : P-value for significance of adding model terms.
            - n2 : Eta-square as effect size (proportion of explained
            variance).
            - np2 : Partial eta-square as partial effect size.

        Notes
        -----        
        The ANOVA table provides information about the significance of 
        each factor and interaction in the model. The type of ANOVA 
        determines how the sum of squares is partitioned among the 
        factors.

        The SAS and also Minitab software uses Type III by default. This
        type is also the only one who gives us a SS and p-value for the 
        Intercept. A discussion on which one to use can be  found here:
        https://stats.stackexchange.com/a/93031

        A nice conclusion about the differences between the types:
        - Typ-I: We choose the most "important" independent variable and 
        it will receive the maximum amount of variation possible.
        - Typ-II: We ignore the shared variation: no interaction is
        assumed. If this is true, the Type II Sums of Squares are
        statistically more powerful. However if in reality there is an
        interaction effect, the model will be wrong and there will be a
        problem in the conclusions of the analysis.
        - Typ-III: If there is an interaction effect and we are looking 
        for an “equal” split between the independent variables, 
        Type-III should be used.

        source:
        https://towardsdatascience.com/anovas-three-types-of-estimating-sums-of-squares-don-t-make-the-wrong-choice-91107c77a27a
        """
        typs = ('I', 'II', 'III')
        assert typ in typs, f'typ must be one of {typs}'

        anova = pd.DataFrame()
        if all(model.pvalues.isna()):
            warnings.warn(
                'ANOVA table could not be calculated because the model is '
                'underdetermined.')
            return anova

        try:
            anova = sm.stats.anova_lm(model, typ=typ)
        except ValueError as error:
            for other_typ in (t for t in typs if t != typ):
                warnings.warn(
                    f'ANOVA table could not be computed with type {typ}. '
                    f'Using type {other_typ} instead.',
                    RuntimeWarning)
                try:
                    anova = sm.stats.anova_lm(model, typ=other_typ)
                except ValueError:
                    continue
                finally:
                    if anova.empty:
                        raise error
        anova = anova.rename(
            columns={'df': 'DF', 'sum_sq': 'SS', 'PR(>F)': 'p'})

        anova['DF'] = anova['DF'].astype(int)
        anova['MS'] = anova['SS']/anova['DF']
        anova['n2'] = anova['SS'] / anova['SS'].sum()
        anova.index.name = ANOVA.SOURCE
        anova.columns.name = f'Typ-{typ}'
        return anova[ANOVA.TABLE_COLNAMES]

def terms_probability(model: RegressionResultsWrapper) -> 'Series[float]':
    """Compute the p-values for the terms in a regression model using a
    ANOVA typ-III table.

    Parameters
    ----------
    model : RegressionResultsWrapper
        The regression model to compute the p-values for.

    Returns
    -------
    Series[float]
        A Series containing the p-values for each term in the model. If 
        the ANOVA table could not be calculated, the p-values will be
        set to NaN.
    
    Notes
    -----
    ANOVA typ III table is used, because it is the only one who gives us
    a p-value for the intercept.
    """
    anova = anova_table(model, typ='III')
    if anova.empty:
        names = model.model.data.design_info.term_names
        p_values = pd.Series({n: np.nan for n in names})
    else:
        p_values = anova['p'].iloc[:-1]
    return p_values



import sys
import pytest

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from pytest import approx
from typing import Generator
from pathlib import Path
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from matplotlib.colors import to_hex
from matplotlib.patches import Patch

sys.path.append(Path(__file__).parent.resolve()) # type: ignore

from daspi import LINE
from daspi import COLOR
from daspi.plotlib.plotter import *


class TestStripe:
    
    class ConcreteStripe(Stripe):
        @property
        def handle(self) -> Patch:
            return Patch()
            
        def __call__(self, ax, **kwds) -> None:
            pass

    @pytest.fixture
    def stripe(self) -> ConcreteStripe:
        return self.ConcreteStripe(
            label='test',
            orientation='horizontal', 
            position=1.0,
            width=0.5
        )

    def test_init_defaults(self, stripe: ConcreteStripe) -> None:
        assert stripe.color == COLOR.STRIPE
        assert stripe.lower_limit == 0.0
        assert stripe.upper_limit == 1.0
        assert stripe.zorder == 0.7
        assert stripe.show_position == False

    def test_decimals_property(self) -> None:
        stripe = self.ConcreteStripe('test', 0.1, 1.0, 'vertical')
        assert stripe.decimals == 4
        
        stripe = self.ConcreteStripe('test', 3.0, 1.0, 'vertical')
        assert stripe.decimals == 3
        
        stripe = self.ConcreteStripe('test', 25.0, 1.0, 'vertical')
        assert stripe.decimals == 2
        
        stripe = self.ConcreteStripe('test', 500.0, 1.0, 'vertical')
        assert stripe.decimals == 1
        
        stripe = self.ConcreteStripe('test', 10000.0, 1.0, 'vertical')
        assert stripe.decimals == 0

    def test_determine_decimals(self) -> None:
        assert Stripe.determine_decimals(0.1) == 4
        assert Stripe.determine_decimals(2.5) == 3
        assert Stripe.determine_decimals(25.0) == 2
        assert Stripe.determine_decimals(500.0) == 1
        assert Stripe.determine_decimals(10000.0) == 0

    def test_label_with_position(self) -> None:
        stripe = self.ConcreteStripe(
            'test', 1.234, 1.0, 'horizontal', show_position=True)
        assert stripe.label == '$test=1.234$'
        
        stripe = self.ConcreteStripe('test', 10000, 1.0, 'horizontal', show_position=True)
        assert stripe.label == '$test=10000$'

    def test_label_without_position(self, stripe: ConcreteStripe) -> None:
        assert stripe.label == '$test$'

    def test_identity(self, stripe: ConcreteStripe) -> None:
        expected = f'$test$_{COLOR.STRIPE}'
        assert stripe.identity == expected


class TestStripeLine:

    @pytest.fixture
    def stripe(self) -> StripeLine:
        return StripeLine(
            label='test',
            orientation='horizontal',
            position=1.0,
            width=2.0
        )
    
    @pytest.fixture
    def ax(self) -> Axes:
        return plt.subplots(1, 1)[1]

    def test_init_defaults(self, stripe: StripeLine) -> None:
        assert stripe.color == COLOR.STRIPE
        assert stripe.lower_limit == 0.0
        assert stripe.upper_limit == 1.0
        assert stripe.zorder == 0.7
        assert stripe.show_position == False
        assert stripe.linestyle == LINE.DASHED

    def test_init_custom_values(self) -> None:
        stripe = StripeLine(
            label='custom',
            orientation='vertical',
            position=2.0,
            color='red',
            lower_limit=0.5,
            upper_limit=1.5,
            zorder=1.0,
            show_position=True,
            linestyle='solid',
            width=3.0
        )
        assert stripe.color == 'red'
        assert stripe.lower_limit == 0.5
        assert stripe.upper_limit == 1.5
        assert stripe.zorder == 1.0
        assert stripe.show_position == True
        assert stripe.linestyle == 'solid'
        assert stripe.width == 3.0
    
    def test_init_priorities(self) -> None:
        stripe = StripeLine(
            label='custom',
            orientation='vertical',
            position=2.0,
            width=1,
            color='red',
            linestyle='solid',)
        assert stripe.width == 1
        assert stripe.color == 'red'
        assert stripe.linestyle == 'solid'
        
        stripe = StripeLine(
            label='custom',
            orientation='vertical',
            position=2.0,
            width=1,
            linewidth=2,
            color='red',
            linestyle='solid',)
        assert stripe.width == 2
        assert stripe.color == 'red'
        assert stripe.linestyle == 'solid'
        stripe = StripeLine(
            label='custom',
            orientation='vertical',
            position=2.0,
            width=1,
            linewidth=2,
            lw=3,
            color='red',
            c='blue',
            linestyle='solid',
            ls='dashed')
        assert stripe.width == 3
        assert stripe.color == 'blue'
        assert stripe.linestyle == 'dashed'

    def test_init_with_kwds(self) -> None:
        stripe = StripeLine(
            label='kwds',
            orientation='horizontal',
            position=1.0,
            ls='dotted',
            lw=4.0,
            c='blue'
        )
        assert stripe.linestyle == 'dotted'
        assert stripe.width == 4.0
        assert stripe.color == 'blue'
    
    def test_call_idempotent(self, stripe: StripeLine, ax: Axes) -> None:
        assert not ax in stripe._axes
        n_axes = len(stripe._axes)
        stripe(ax)
        n_lines = len(ax.lines)
        assert ax in stripe._axes
        assert len(stripe._axes) == n_axes + 1
        stripe(ax)
        assert len(ax.lines) == n_lines
        assert len(stripe._axes) == n_axes + 1

    def test_call_horizontal(self, stripe: StripeLine, ax: Axes) -> None:
        stripe(ax)
        assert ax in stripe._axes
        assert len(ax.lines) == 1
        line = ax.lines[0]
        assert line.get_ydata()[0] == stripe.position # type: ignore
        assert line.get_color() == stripe.color
        assert line.get_linewidth() == stripe.width
        assert line.get_zorder() == stripe.zorder

    def test_call_vertical(self, ax: Axes) -> None:
        stripe = StripeLine(
            label='vertical',
            orientation='vertical',
            position=2.0,
            color='red')
        assert not ax in stripe._axes
        n_lines = len(ax.lines)
        stripe(ax)
        assert len(ax.lines) == n_lines + 1
        assert ax in stripe._axes
        line = ax.lines[-1]
        assert line.get_xdata()[0] == stripe.position # type: ignore
        assert line.get_color() == stripe.color
        assert line.get_linewidth() == stripe.width
        assert line.get_zorder() == stripe.zorder

    def test_handle_properties(self, stripe: StripeLine) -> None:
        handle = stripe.handle
        assert isinstance(handle, Line2D)
        assert handle.get_color() == COLOR.STRIPE
        assert handle.get_linewidth() == stripe.width
        assert handle.get_linestyle() == '--'
        assert handle.get_markersize() == 0
        assert len(handle.get_xdata()) == 0 # type: ignore
        assert len(handle.get_ydata()) == 0 # type: ignore

        stripe = StripeLine(
            label='test',
            orientation='horizontal', 
            position=1.0,
            width=3.0,
            color='red',
            linestyle=':')
        handle = stripe.handle
        assert handle.get_color() == 'red'
        assert handle.get_linewidth() == 3.0
        assert handle.get_linestyle() == ':'



class TestStripeSpan:

    @pytest.fixture
    def ax(self) -> Axes:
        return plt.subplots(1, 1)[1]

    def test_init_with_position_width(self) -> None:
        stripe = StripeSpan(
            label='test',
            orientation='horizontal',
            position=1.0,
            width=2.0)
        assert stripe.lower_position == 0.0
        assert stripe.upper_position == 2.0
        assert stripe.position == 1.0
        assert stripe.width == 2.0
        assert stripe.color == COLOR.STRIPE
        assert stripe.alpha == COLOR.CI_ALPHA
        assert stripe.lower_limit == 0.0
        assert stripe.upper_limit == 1.0
        assert stripe.zorder == 0.7
        assert stripe.border_linewidth == 0

    def test_init_with_lower_upper_position(self) -> None:
        stripe = StripeSpan(
            label='test',
            orientation='vertical',
            lower_position=0.0,
            upper_position=2.0)
        assert stripe.lower_position == 0.0
        assert stripe.upper_position == 2.0
        assert stripe.position == 1.0
        assert stripe.width == 2.0

    def test_init_invalid_parameters(self) -> None:
        with pytest.raises(AssertionError):
            StripeSpan(
                label='test',
                orientation='horizontal',
                position=1.0,
                width=2.0,
                lower_position=0.0,
                upper_position=2.0)

        with pytest.raises(AssertionError):
            StripeSpan(
                label='test',
                orientation='horizontal')
            
        with pytest.raises(AssertionError):
            StripeSpan(
                label='test',
                orientation='horizontal',
                position=1.0,
                lower_position=0.0,
                upper_position=2.0)

        with pytest.raises(AssertionError):
            StripeSpan(
                label='test',
                orientation='horizontal',
                position=1.0,
                width=2.0,
                upper_position=2.0)

    def test_handle_property(self) -> None:
        stripe = StripeSpan(
            label='test',
            orientation='horizontal',
            position=1.0,
            width=2.0)
        handle = stripe.handle
        assert isinstance(handle, Patch)
        assert handle.get_alpha() == COLOR.CI_ALPHA
        assert to_hex(handle.get_facecolor()) == COLOR.STRIPE.lower()
        
        stripe = StripeSpan(
            label='test',
            orientation='horizontal',
            position=1.0,
            width=2.0,
            color=COLOR.PALETTE[0],
            alpha=0.123)
        handle = stripe.handle
        assert isinstance(handle, Patch)
        assert handle.get_alpha() == 0.123
        assert to_hex(handle.get_facecolor()) == to_hex(COLOR.PALETTE[0]).lower()

    def test_call_horizontal(self, ax: Axes) -> None:
        stripe = StripeSpan(
            label='test',
            orientation='horizontal',
            position=1.0,
            width=2.0,
            border_linewidth=1.0)
        stripe(ax)
        assert ax in stripe._axes
        assert len(ax.patches) == 1
        patch = ax.patches[0]
        assert patch.get_alpha() == COLOR.CI_ALPHA
        assert patch.get_linewidth() == 1.0
        assert patch.get_zorder() == 0.7

    def test_call_vertical(self, ax: Axes) -> None:
        stripe = StripeSpan(
            label='test',
            orientation='vertical',
            position=1.0,
            width=2.0)
        stripe(ax)
        assert ax in stripe._axes
        assert len(ax.patches) == 1

    def test_call_idempotent(self, ax: Axes) -> None:
        stripe = StripeSpan(
            label='test',
            orientation='horizontal',
            position=1.0,
            width=2.0)
        stripe(ax)
        n_patches = len(ax.patches)
        stripe(ax)
        assert len(ax.patches) == n_patches

    def test_custom_styling(self, ax: Axes) -> None:
        stripe = StripeSpan(
            label='test',
            orientation='horizontal',
            position=1.0,
            width=2.0,
            color='red',
            alpha=0.3,
            zorder=2.0,
            border_linewidth=2.0)
        stripe(ax)
        patch = ax.patches[0]
        assert patch.get_facecolor()[:3] == (1.0, 0.0, 0.0)
        assert patch.get_alpha() == 0.3
        assert patch.get_zorder() == 2.0
        assert patch.get_linewidth() == 2.0



import sys
import pytest
import numpy as np
from pathlib import Path
from matplotlib.ticker import FuncFormatter
from matplotlib.ticker import PercentFormatter
from matplotlib.ticker import StrMethodFormatter
from unittest.mock import Mock, patch


sys.path.append(Path(__file__).parent.resolve()) # type: ignore

from daspi import AxesFacets
from daspi import flat_unique
from daspi.plotlib.facets import LabelFacets


def test_flat_unique_numpy() -> None:
    nested = np.array([[1, 1], [2, 3]])
    result = flat_unique(nested)
    assert result == [1, 2, 3]

    nested = np.array([[1, 1, 1], [2, 2, 3], [2, 2, 3]])
    result = flat_unique(nested)
    assert result == [1, 2, 3]

    nested = np.array([[2, 2, 3], [2, 2, 3], [1, 1, 1]])
    result = flat_unique(nested)
    assert result == [2, 3, 1]

def test_flat_unique_list() -> None:
    nested = [[1, 1], [2, 3]]
    result = flat_unique(nested)
    assert result == [1, 2, 3]

    nested = [[1, 1, 1], [2, 2, 3], [2, 2, 3]]
    result = flat_unique(nested)
    assert result == [1, 2, 3]

    nested = [[2, 2, 3], [2, 2, 3], [1, 1, 1]]
    result = flat_unique(nested)
    assert result == [2, 3, 1]

def test_flat_unique_empty() -> None:
    result = flat_unique([])
    assert result == []


class TestAxesFacets:

    @pytest.fixture
    def rc(self) -> AxesFacets:
        axes = AxesFacets(
            nrows=2, ncols=2, sharex='col', sharey='row', 
            width_ratios=[4, 1], height_ratios=[1, 4])
        return axes
    
    @pytest.fixture
    def s_mosaic(self) -> AxesFacets:
        mosaic="""
            AAA
            BBC
            BBC"""
        axes = AxesFacets(
            mosaic=mosaic, sharex='col', sharey='row', 
            width_ratios=[2, 2, 1], height_ratios=[1, 2, 2])
        return axes
    
    @pytest.fixture
    def l_mosaic(self) -> AxesFacets:
        mosaic=[
            ['Horizontal', 'Horizontal', '.'],
            ['Vertical', 'Qubic big', 'Qubic big'],
            ['Vertical', 'Qubic big', 'Qubic big']]
        axes = AxesFacets(
            mosaic=mosaic, sharex='col', sharey='row', 
            width_ratios=[2, 2, 1], height_ratios=[1, 2, 2])
        return axes
    
    def test_axes_grid_rc(self,rc: AxesFacets) -> None:
        assert rc.shape == (2, 2)
        assert rc.nrows == 2
        assert rc.ncols == 2
        assert len(rc) == 4

    def test_axes_grid_string(self, s_mosaic: AxesFacets) -> None:
        assert s_mosaic.shape == (3, 3)
        assert s_mosaic.nrows == 3
        assert s_mosaic.ncols == 3
        assert len(s_mosaic) == 3
        assert all(s_mosaic[0] == a for a in s_mosaic[0, :])
        assert not any(s_mosaic[0] == a for a in s_mosaic[1:, :].flat)
        assert all(s_mosaic[1] == a for a in s_mosaic[1:, :2].flat)
        assert not any(s_mosaic[1] == a for a in s_mosaic[1:, 2])
        assert all(s_mosaic[2] == a for a in s_mosaic[1:, 2])

    def test_axes_grid_lists(self, l_mosaic: AxesFacets) -> None:
        assert l_mosaic.shape == (3, 3)
        assert l_mosaic.nrows == 3
        assert l_mosaic.ncols == 3
        assert l_mosaic[0, -1] is None
        assert len(l_mosaic) == 3
        assert all(l_mosaic[0] == a for a in l_mosaic[0, :2])
        assert l_mosaic[0] != l_mosaic[0, -1]
        assert l_mosaic[0] not in l_mosaic[1:, :].flat
        assert all(l_mosaic[1] == a for a in l_mosaic[1:, 0])
        assert l_mosaic[1] not in l_mosaic[1:, 1:].flat
        assert all(l_mosaic[2] == a for a in l_mosaic[1:, 1:].flat)

    def test_iteration(self, l_mosaic: AxesFacets) -> None:
        assert l_mosaic.ax is None
        assert next(iter(l_mosaic)) == l_mosaic[0]
        for i, ax in enumerate(l_mosaic):
            assert l_mosaic.ax == ax
            assert ax == l_mosaic[i]


class TestLabelFacets:
    
    @pytest.fixture
    def simple_axes(self) -> AxesFacets:
        return AxesFacets(nrows=1, ncols=2)
    
    def percentage_formatter(self, x, pos):
        """Format tick values as percentages."""
        return f'{x:.1f}%'
    
    def test_label_facets_initialization(self, simple_axes: AxesFacets) -> None:
        """Test that LabelFacets initializes correctly with new parameters."""
        label_facets = LabelFacets(
            simple_axes,
            xlabel='Test X Label',
            ylabel='Test Y Label',
            xlabel_formatter=FuncFormatter(self.percentage_formatter),
            ylabel_formatter=FuncFormatter(self.percentage_formatter),
            xlabel_angle=45,
            ylabel_angle=30
        )
        
        assert label_facets.xlabel == 'Test X Label'
        assert label_facets.ylabel == 'Test Y Label'
        assert label_facets.xlabel_formatter is not None
        assert label_facets.ylabel_formatter is not None
        assert label_facets.xlabel_angle == 45
        assert label_facets.ylabel_angle == 30
    
    def test_label_facets_default_values(self, simple_axes: AxesFacets) -> None:
        """Test that LabelFacets has correct default values for new parameters."""
        label_facets = LabelFacets(simple_axes)
        
        assert label_facets.xlabel_formatter is None
        assert label_facets.ylabel_formatter is None
        assert label_facets.xlabel_angle == 0
        assert label_facets.ylabel_angle == 0
    
    def test_label_facets_draw_with_formatters_and_angles(self, simple_axes: AxesFacets) -> None:
        """Test that formatters and angles are applied correctly when drawing."""
        # Add some data to the axes
        for ax in simple_axes:
            ax.plot([1, 2, 3], [10, 20, 30])
        
        label_facets = LabelFacets(
            simple_axes,
            xlabel_formatter=FuncFormatter(self.percentage_formatter),
            ylabel_formatter=FuncFormatter(self.percentage_formatter),
            xlabel_angle=45,
            ylabel_angle=15
        )
        
        # Draw the labels
        label_facets.draw()
        
        # Verify that angles were applied
        for ax in simple_axes:
            x_tick_labels = ax.get_xticklabels()
            y_tick_labels = ax.get_yticklabels()
            
            if x_tick_labels:
                assert x_tick_labels[0].get_rotation() == 45
            if y_tick_labels:
                assert y_tick_labels[0].get_rotation() == 15
    
    def test_label_facets_alignment_features(self, simple_axes: AxesFacets) -> None:
        """Test that alignment features work correctly."""
        # Add some data to the axes
        for ax in simple_axes:
            ax.plot([1, 2, 3], [10, 20, 30])
        
        label_facets = LabelFacets(
            simple_axes,
            xlabel_align='left',
            ylabel_align='top',
            xlabel_angle=30,
            ylabel_angle=45
        )
        
        # Draw the labels
        label_facets.draw()
        
        # Verify alignment was applied
        for ax in simple_axes:
            x_tick_labels = ax.get_xticklabels()
            y_tick_labels = ax.get_yticklabels()
            
            if x_tick_labels:
                assert x_tick_labels[0].get_horizontalalignment() == 'left'
                assert x_tick_labels[0].get_rotation() == 30
            if y_tick_labels:
                assert y_tick_labels[0].get_verticalalignment() == 'top'
                assert y_tick_labels[0].get_rotation() == 45
    
    def test_label_facets_alignment_with_default_values(self, simple_axes: AxesFacets) -> None:
        """Test alignment with default center values."""
        # Add some data to the axes
        for ax in simple_axes:
            ax.plot([1, 2, 3], [10, 20, 30])
        
        label_facets = LabelFacets(simple_axes)
        
        # Verify default alignment values
        assert label_facets.xlabel_align == 'center'
        assert label_facets.ylabel_align == 'center'
        
        # Draw and verify no changes to alignment when center (default matplotlib behavior)
        label_facets.draw()


class TestLabelFacetsFormatting:
    """Test the new formatting features of LabelFacets."""

    @pytest.fixture
    def axes_with_data(self) -> AxesFacets:
        """Create AxesFacets with some sample data."""
        axes = AxesFacets(nrows=1, ncols=1)
        # Add sample data to create ticks
        # Access the actual axes object from the numpy array
        actual_ax = axes.axes.flat[0]  # This gets the actual matplotlib Axes
        actual_ax.plot([1, 2, 3, 4, 5], [10.5, 20.3, 30.7, 40.2, 50.9])
        return axes

    def test_prepare_formatter_with_none(self, axes_with_data: AxesFacets) -> None:
        """Test _prepare_formatter with None input."""
        label_facets = LabelFacets(axes_with_data)
        result = label_facets._prepare_formatter(None)
        assert result is None

    def test_prepare_formatter_with_existing_formatter(self, axes_with_data: AxesFacets) -> None:
        """Test _prepare_formatter with existing Formatter."""
        existing_formatter = PercentFormatter()
        label_facets = LabelFacets(axes_with_data)
        result = label_facets._prepare_formatter(existing_formatter)
        assert result is existing_formatter

    def test_prepare_formatter_with_single_arg_callable(self, axes_with_data: AxesFacets) -> None:
        """Test _prepare_formatter with single-argument callable."""
        def single_arg_formatter(x):
            return f"{x:.1f}°C"
        
        label_facets = LabelFacets(axes_with_data)
        result = label_facets._prepare_formatter(single_arg_formatter)
        
        assert isinstance(result, FuncFormatter)
        # Test that the wrapped formatter works
        formatted_value = result(25.6, 0)
        assert formatted_value == "25.6°C"

    def test_prepare_formatter_with_two_arg_callable(self, axes_with_data: AxesFacets) -> None:
        """Test _prepare_formatter with two-argument callable."""
        def two_arg_formatter(x, pos):
            return f"#{pos}: {x:.2f}"
        
        label_facets = LabelFacets(axes_with_data)
        result = label_facets._prepare_formatter(two_arg_formatter)
        
        assert isinstance(result, FuncFormatter)
        # Test that the formatter works
        formatted_value = result(25.6, 2)
        assert formatted_value == "#2: 25.60"

    def test_prepare_formatter_with_string_positional(self, axes_with_data: AxesFacets) -> None:
        """Test _prepare_formatter with positional string formats."""
        test_cases = [
            ('{:.2e}', 1234.567, '1.23e+03'),
            ('{:.1f}', 1234.567, '1234.6'),
            ('{:.0f}', 1234.567, '1235'),
            ('${:.2f}', 1234.567, '$1234.57'),
            ('{:.2%}', 0.1234, '12.34%'),
            ('${:.0f}k', 1234.567, '$1235k'),
        ]
        
        label_facets = LabelFacets(axes_with_data)
        
        for format_string, test_value, expected in test_cases:
            result = label_facets._prepare_formatter(format_string)
            assert isinstance(result, FuncFormatter)
            formatted_value = result(test_value, 0)
            assert formatted_value == expected, f"Format '{format_string}' failed: got '{formatted_value}', expected '{expected}'"

    def test_prepare_formatter_with_string_keyword(self, axes_with_data: AxesFacets) -> None:
        """Test _prepare_formatter with keyword string formats."""
        label_facets = LabelFacets(axes_with_data)
        
        # Test keyword format (should use StrMethodFormatter)
        result = label_facets._prepare_formatter('{x:.2e}')
        assert isinstance(result, StrMethodFormatter)
        
        # Test that it works
        formatted_value = result(1234.567, 0)
        assert formatted_value == '1.23e+03'

    def test_formatter_integration_with_axes(self, axes_with_data: AxesFacets) -> None:
        """Test that formatters are properly applied to axes."""
        def temp_formatter(x):
            return f"{x:.1f}°C"
        
        def pressure_formatter(x):
            return f"{x:.2f} bar"
        
        label_facets = LabelFacets(
            axes_with_data,
            xlabel_formatter=pressure_formatter,
            ylabel_formatter=temp_formatter
        )
        label_facets.draw()
        
        # Verify formatters are applied
        actual_ax = axes_with_data.axes.flat[0]
        x_formatter = actual_ax.xaxis.get_major_formatter()
        y_formatter = actual_ax.yaxis.get_major_formatter()
        
        assert isinstance(x_formatter, FuncFormatter)
        assert isinstance(y_formatter, FuncFormatter)

    def test_rotation_angles_applied(self, axes_with_data: AxesFacets) -> None:
        """Test that rotation angles are properly applied."""
        label_facets = LabelFacets(
            axes_with_data,
            xlabel_angle=45,
            ylabel_angle=90
        )
        label_facets.draw()
        
        # Check that rotation is applied to tick labels
        actual_ax = axes_with_data.axes.flat[0]
        x_labels = actual_ax.get_xticklabels()
        y_labels = actual_ax.get_yticklabels()
        
        if x_labels:
            assert x_labels[0].get_rotation() == 45
        if y_labels:
            assert y_labels[0].get_rotation() == 90

    def test_alignment_options_applied(self, axes_with_data: AxesFacets) -> None:
        """Test that alignment options are properly applied."""
        label_facets = LabelFacets(
            axes_with_data,
            xlabel_align='right',
            ylabel_align='top'
        )
        label_facets.draw()
        
        # Check that alignment is applied
        actual_ax = axes_with_data.axes.flat[0]
        x_labels = actual_ax.get_xticklabels()
        y_labels = actual_ax.get_yticklabels()
        
        if x_labels:
            assert x_labels[0].get_horizontalalignment() == 'right'
        if y_labels:
            assert y_labels[0].get_verticalalignment() == 'top'

    def test_rotation_margin_estimation(self, axes_with_data: AxesFacets) -> None:
        """Test the rotation margin estimation."""
        label_facets = LabelFacets(axes_with_data)
        
        # Test with no rotation
        margin_0 = label_facets.estimate_rotation_margin(0)
        assert margin_0 == 0
        
        # Test with small rotation (should be 0)
        margin_small = label_facets.estimate_rotation_margin(4)
        assert margin_small == 0
        
        # Test with 45 degree rotation
        margin_45 = label_facets.estimate_rotation_margin(45, base_margin=20)
        expected_45 = int(20 * np.sin(45 * np.pi / 180))
        assert margin_45 == expected_45
        
        # Test with 90 degree rotation
        margin_90 = label_facets.estimate_rotation_margin(90, base_margin=20)
        expected_90 = int(20 * np.sin(90 * np.pi / 180))
        assert margin_90 == expected_90

    def test_combined_formatting_features(self, axes_with_data: AxesFacets) -> None:
        """Test combining all formatting features together."""
        def custom_formatter(x):
            return f"${x:,.0f}"
        
        label_facets = LabelFacets(
            axes_with_data,
            xlabel='Cost',
            ylabel='Revenue',
            xlabel_formatter=custom_formatter,
            ylabel_formatter=lambda x: f"{x:.1f}%",
            xlabel_angle=30,
            ylabel_angle=15,
            xlabel_align='right',
            ylabel_align='bottom'
        )
        label_facets.draw()
        
        # Verify all features are applied
        actual_ax = axes_with_data.axes.flat[0]
        x_formatter = actual_ax.xaxis.get_major_formatter()
        y_formatter = actual_ax.yaxis.get_major_formatter()
        
        assert isinstance(x_formatter, FuncFormatter)
        assert isinstance(y_formatter, FuncFormatter)
        
        x_labels = actual_ax.get_xticklabels()
        y_labels = actual_ax.get_yticklabels()
        
        if x_labels:
            assert x_labels[0].get_rotation() == 30
            assert x_labels[0].get_horizontalalignment() == 'right'
        if y_labels:
            assert y_labels[0].get_rotation() == 15
            assert y_labels[0].get_verticalalignment() == 'bottom'



import warnings

from os import environ
from typing import Dict
from typing import Literal
from datetime import date


__all__ = ['STR']


class _String_:

    anderson_darling: Dict[str, str] = {
        'en': 'Anderson-Darling',
        'de': 'Anderson-Darling',
        'fr': 'Anderson-Darling'}

    ok: Dict[str, str] = { 
        'en': 'OK',
        'de': 'IO',
        'fr': 'OK'}

    nok: Dict[str, str] = { 
        'en': 'NOK',
        'de': 'NIO',
        'fr': 'NOK'}
    
    accepted: Dict[str, str] = {
        'en': 'accepted',
        'de': 'akzeptiert',
        'fr': 'accepté'}
    
    rejected: Dict[str, str] = {
        'en': 'rejected',
        'de': 'abgelehnt',
        'fr': 'rejeté'}
    
    borderline: Dict[str, str] = {
        'en': 'borderline',
        'de': 'grenzwertig',
        'fr': 'limite'}

    lsl: Dict[str, str] = { 
        'en': 'LSL',
        'de': 'USG',
        'fr': 'LSL'}
    
    usl: Dict[str, str] = {
        'en': 'USL',
        'de': 'OSG',
        'fr': 'USL'}
    
    lcl: Dict[str, str] = { 
        'en': 'LCL',
        'de': 'UEG',
        'fr': 'LCL'}
    
    ucl: Dict[str, str] = {
        'en': 'UCL',
        'de': 'OEG',
        'fr': 'UCL'}
    
    excess: Dict[str, str] = {
        'en': 'excess',
        'de': 'Exzess',
        'fr': 'excès'}
    
    skew: Dict[str, str] = {
        'en': 'skew',
        'de': 'Schiefe',
        'fr': 'asymétrie'}
    
    kde_ax_label: Dict[str, str] = {
        'en': 'Estimated kernel density',
        'de': 'Geschätzte Kerndichte',
        'fr': 'Densité de noyau estimée'}

    stripes: Dict[str, str] = {
        'en': 'Lines',
        'de': 'Linien',
        'fr': 'Lignes'}

    ci: Dict[str, str] = {
        'en': 'CI',
        'de': 'KI',
        'fr': 'IC'}
    
    formula: Dict[str, str] = {
        'en': 'formula',
        'de': 'Formel',
        'fr': 'formule'}
    
    effects_label: Dict[str, str] = {
        'en': 'Standardized effect',
        'de': 'Standardisierter Effekt',
        'fr': 'Effet standardisé'}
    
    ss_label: Dict[str, str] = {
        'en': 'Sum of Squares',
        'de': 'Summenquadrate',
        'fr': 'Somme des carrés'}
    
    data_range: Dict[str, str] = {
        'en': 'Range',
        'de': 'Spannweite',
        'fr': 'Plage des données'}
    
    paramcharts_fig_title: Dict[str, str] = {
        'en': 'Parameter Analysis',
        'de': 'Parameter Analyse',
        'fr': 'Analyse des paramètres'}
    
    paramcharts_sub_title: Dict[str, str] = {
        'en': 'Relative importance of parameters',
        'de': 'Relative Wichtigkeit der Parameter',
        'fr': 'Importance relative des paramètres'}
    
    paramcharts_feature_label: Dict[str, str] = {
        'en': 'Parameter',
        'de': 'Parameter',
        'fr': 'Paramètre'}
    
    residcharts_fig_title: Dict[str, str] = {
        'en': 'Residuals analysis',
        'de': 'Residuen Analyse',
        'fr': 'Analyse des résidus'}
    
    resid_name: Dict[str, str] = {
        'en': 'Residuals',
        'de': 'Residuen',
        'fr': 'Résidus'}
    
    fit: Dict[str, str] = {
        'en': 'Fit',
        'de': 'Anpassung',
        'fr': 'Ajustement'}
    
    charts_flabel_quantiles: Dict[str, str] = {
        'en': 'Std. Normal Distribution quantiles',
        'de': 'Standardnormalverteilung Quantile',
        'fr': 'Quantiles de la distribution normale standard'}
    
    charts_flabel_density: Dict[str, str] = {
        'en': 'Estimated kernel density',
        'de': 'Geschätzte Kerndichte',
        'fr': 'Densité de noyau estimée'}
    
    charts_flabel_predicted: Dict[str, str] = {
        'en': 'Predicted values',
        'de': 'Vorhersage',
        'fr': 'Valeurs prédites'}
    
    charts_flabel_observed: Dict[str, str] = {
        'en': 'Observation order',
        'de': 'Beobachtungsreihenfolge',
        'fr': 'Ordre d\'observation'}
    
    charts_label_alpha_th: Dict[str, str] = {
        'en': r'effect_α\;(α={alpha})',
        'de': r'Effekt_α\;(α={alpha})',
        'fr': r'effet_α\;(α={alpha})'}

    cp: Dict[str, str] = {
        'en': 'Process Capability index Cp',
        'de': 'Prozessfähigkeitsindex Cp',
        'fr': 'Indice de capacité de processus Cp'}

    cpk: Dict[str, str] = {
        'en': 'Adjusted Process Capability index Cpk',
        'de': 'Angepasster Prozessfähigkeitsindex Cpk',
        'fr': 'Indice de capacité de processus ajusté Cpk'}
    
    paircharts_fig_title: Dict[str, str] = {
        'en': 'Pairwise analysis',
        'de': 'Paarweise Analyse',
        'fr': 'Analyse pair à pair'}
    
    paircharts_sub_title: Dict[str, str] = {
        'en': 'Bland-Altman 95 % CI and individual value comparison',
        'de': 'Bland-Altman 95 %-KI und Einzelwertvergleich',
        'fr': 'Bland-Altman 95 % IC et comparaison des valeurs individuelles'}
    
    gstudycharts_fig_title: Dict[str, str] = {
        'en': 'Measurement system analysis',
        'de': 'Messsystemanalyse',
        'fr': 'Analyse du système de mesure'}
    
    gstudycharts_sub_title: Dict[str, str] = {
        'en': 'Gage study type 1',
        'de': 'MSA Typ 1',
        'fr': 'Etude de système de mesure type 1'}
    
    rnrcharts_fig_title: Dict[str, str] = {
        'en': 'Measurement system analysis',
        'de': 'Messsystemanalyse',
        'fr': 'Analyse du système de mesure'}

    rnrcharts_sub_title: Dict[str, str] = {
        'en': 'Repeatability and reproducibility (Gage R&R)',
        'de': 'Wiederholbarkeit und Reproduzierbarkeit (Gage R&R)',
        'fr': 'Répétabilité et reproductibilité (Gage R&R)'}
    
    rnrcharts_spread_proportions: Dict[str, str] = {
        'en': 'Spread proportions',
        'de': 'Streuungsanteile',
        'fr': 'Proportions de dispersion'}
    
    rnrcharts_suitability: Dict[str, str] = {
        'en': 'Suitability index Q',
        'de': 'Eignungskennwert Q',
        'fr': 'Indice d\'adéquation Q'}
    
    lm_table_caption_summary: Dict[str, str] = {
        'en': 'Model summary',
        'de': 'Modellzusammenfassung',
        'fr': 'Résumé du modèle'}
    
    lm_table_caption_statistics: Dict[str, str] = {
        'en': 'Parameter statistics',
        'de': 'Parameterstatistik',
        'fr': 'Statistiques des paramètres'}
    
    lm_table_caption_anova: Dict[str, str] = {
        'en': 'Analysis of variance',
        'de': 'Varianzanalyse',
        'fr': 'Analyse de la variance'}
    
    lm_table_caption_vif: Dict[str, str] = {
        'en': 'Variance inflation factor',
        'de': 'Varianzinflationfaktor',
        'fr': 'Facteur d\'inflation de la variance'}
    
    lm_table_caption_rnr: Dict[str, str] = {
        'en': 'Repeatability and reproducibility (R&R)',
        'de': 'Wiederholbarkeit und Reproduzierbarkeit (R&R)',
        'fr': 'Répétabilité et reproductibilité (R&R)'}
    
    lm_table_rnr_source: Dict[str, str] = {
        'en': 'Source',
        'de': 'Quelle',
        'fr': 'Source'}
    
    lm_table_caption_ref_gages: Dict[str, str] = {
        'en': 'Reference analysis',
        'de': 'Analyse der Referenzen',
        'fr': 'Analyse des références'}

    lm_table_caption_ms_uncertainty: Dict[str, str] = {
        'en': 'Measurement uncertainty budget measuring system',
        'de': 'Messunsicherheitsbudget Messsystem',
        'fr': 'Système de mesure du budget d\'incertitude de mesure'}
    
    lm_table_caption_mp_uncertainty: Dict[str, str] = {
        'en': 'Measurement uncertainty budget measuring process',
        'de': 'Messunsicherheitsbudget Messprozess',
        'fr': 'Budget d\'incertitude de mesure de la méthode de mesure'}
    
    lm_table_caption_capabilities: Dict[str, str] = {
        'en': 'Gage capability',
        'de': 'Messsystemfähigkeit',
        'fr': 'Capacité du système de mesure'}

    _language_: Literal['en', 'de', 'fr'] = 'en'
    _username_: str = environ.get('USERNAME', 'user')

    @property
    def TODAY(self) -> str:
        return date.today().strftime('%Y-%m-%d')
    
    @property
    def LANGUAGE(self) -> Literal['en', 'de', 'fr']:
        """Language (abbreviation) in which the strings should be
        rendered"""
        return self._language_
    @LANGUAGE.setter
    def LANGUAGE(self, language: Literal['en', 'de', 'fr']) -> None:
        assert language in ('en', 'de', 'fr')
        self._language_ = language
    
    @property
    def USERNAME(self) -> str:
        """Username reflected in the charts in the info text, defaults 
        to username from the environment variable."""
        return self._username_
    @USERNAME.setter
    def USERNAME(self, username: str) -> None:
        self._username_ = username
    
    def __getitem__(self, item: str) -> str | Literal['']:
        _string = ''
        try:
            _string = getattr(self, item)[self.LANGUAGE]
        except AttributeError:
            warnings.warn(f'No string found for {item}!')
        except KeyError:
            warnings.warn(
                f'No string found for {item} and language {self.LANGUAGE}!')
        return _string
STR = _String_()
